# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse

from yinyang.src.base.Exitcodes import ERR_USAGE
from yinyang.src.base.Version import VERSION, COMMIT


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        print("usage:" + self.usage, flush=True)
        self.exit(
            ERR_USAGE,
            "error: %s\nFor a listing of options, use %s --help.\n"
            % (message, self.prog),
        )


def add_common_args(parser, rootpath, current_dir):
    parser.add_argument(
        "SOLVER_CLIS",
        metavar="solver_clis",
    )
    parser.add_argument(
        "PATH_TO_SEEDS",
        nargs="+",
        metavar="seed_file/seed_folder",
    )
    parser.add_argument(
        "-l",
        "--logfolder",
        metavar="path_path_to_folder",
        default=current_dir + "/logs",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-t",
        "--timeout",
        default=8,
        metavar="secs",
        type=int,
    )
    parser.add_argument(
        "-b",
        "--bugsfolder",
        metavar="path_to_folder",
        default=current_dir + "/bugs",
    )
    parser.add_argument(
        "-s",
        "--scratchfolder",
        metavar="path_to_folder",
        default=current_dir + "/scratch",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
    )
    parser.add_argument(
        "-n",
        "--no-log",
        action="store_true",
    )
    parser.add_argument(
        "-L",
        "--file-size-limit",
        metavar="num_bytes",
        default=100000,
        type=int,
    )


def add_opfuzz_args(parser, rootpath, current_dir):
    parser.add_argument(
        "-i",
        "--iterations",
        default=300,
        metavar="<N>",
        type=int,
    )
    parser.add_argument(
        "-m",
        "--modulo",
        default=2,
        metavar="<N>",
        type=int,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="opfuzz " + VERSION + " " + "(" + COMMIT + ")",
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="path_to_file",
        default=rootpath + "/yinyang/config/operator_mutations.txt",
    )


def add_typefuzz_args(parser, rootpath, current_dir):
    parser.add_argument(
        "-i",
        "--iterations",
        default=300,
        metavar="<N>",
        type=int,
    )
    parser.add_argument(
        "-m",
        "--modulo",
        default=2,
        metavar="<N>",
        type=int,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="typefuzz " + VERSION,
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="path_to_file",
        default=rootpath + "/yinyang/config/typefuzz_config.txt",
    )


def add_yinyang_args(parser, rootpath, current_dir):
    parser.add_argument(
        "-o",
        "--oracle",
        default="sat",
        metavar="{sat,unsat}",
        type=str,
    )
    parser.add_argument(
        "-i",
        "--iterations",
        default=30,
        metavar="<N>",
        type=int,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="yinyang " + VERSION + " " + "(" + COMMIT + ")",
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="path_to_file",
        default=rootpath + "/yinyang/config/fusion_functions.txt",
    )


def build_opfuzz_parser(rootpath, current_dir, usage):
    parser = ArgumentParser(
        description="",
        usage=usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,
    )
    add_common_args(parser, rootpath, current_dir)
    add_opfuzz_args(parser, rootpath, current_dir)

    return parser


def build_typefuzz_parser(rootpath, current_dir, usage):
    parser = ArgumentParser(
        description="",
        usage=usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,
    )
    add_common_args(parser, current_dir, rootpath)
    add_typefuzz_args(parser, current_dir, rootpath)

    return parser


def build_yinyang_parser(rootpath, current_dir, usage):
    parser = ArgumentParser(
        description="",
        usage=usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,
    )
    add_common_args(parser, current_dir, rootpath)
    add_yinyang_args(parser, current_dir, rootpath)

    return parser



# MIT License
#
# Copyright (c) [2020 - 2020] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import subprocess
from enum import Enum

from yinyang.src.base.Exitcodes import ERR_USAGE


class SolverQueryResult(Enum):
    """
    Enum storing the result of a single solver check-sat query.
    """

    SAT = 0  # solver query returns "sat"
    UNSAT = 1  # solver query returns "unsat"
    UNKNOWN = 2  # solver query reports "unknown"


def sr2str(sol_res):
    if sol_res == SolverQueryResult.SAT:
        return "sat"
    if sol_res == SolverQueryResult.UNSAT:
        return "unsat"
    if sol_res == SolverQueryResult.UNKNOWN:
        return "unknown"


class SolverResult:
    """
    Class to store the result of multiple solver check-sat queries.
    :lst a list of multiple "SolverQueryResult" items
    """

    def __init__(self, result=None):
        self.lst = []
        if result:
            self.lst.append(result)

    def append(self, result):
        self.lst.append(result)

    def equals(self, rhs):
        if type(rhs) == SolverQueryResult:
            return len(self.lst) == 1 and self.lst[0] == rhs
        elif type(rhs) == SolverResult:
            if len(self.lst) != len(rhs.lst):
                return False
            for index in range(0, len(self.lst)):
                if (
                    self.lst[index] != SolverQueryResult.UNKNOWN
                    and rhs.lst[index] != SolverQueryResult.UNKNOWN
                    and self.lst[index] != rhs.lst[index]
                ):
                    return False
            return True
        else:
            return False

    def __str__(self):
        s = sr2str(self.lst[0])
        for res in self.lst[1:]:
            s += "\n" + sr2str(res)
        return s


class Solver:
    def __init__(self, cil):
        self.cil = cil

    def solve(self, file, timeout, debug=False):
        try:
            cmd = list(filter(None, self.cil.split(" "))) + [file]
            if debug:
                print("cmd: " + " ".join(cmd), flush=True)
            output = subprocess.run(
                cmd,
                timeout=timeout,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=False,
            )

        except subprocess.TimeoutExpired as te:
            if te.stdout and te.stderr:
                stdout = te.stdout.decode()
                stderr = te.stderr.decode()
            else:
                stdout = ""
                stderr = ""
            return stdout, stderr, 137

        except ValueError:
            stdout = ""
            stderr = ""
            return stdout, stderr, 0

        except FileNotFoundError:
            print('error: solver "' + cmd[0] + '" not found', flush=True)
            exit(ERR_USAGE)

        stdout = output.stdout.decode()
        stderr = output.stderr.decode()
        returncode = output.returncode

        if debug:
            print("output: " + stdout + "\n" + stderr)

        return stdout, stderr, returncode



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random

from yinyang.src.mutators.Mutator import Mutator
from yinyang.src.mutators.GenTypeAwareMutation.Operator import (
    Operator, handle_parametric_op, handle_non_parametric_op
)
from yinyang.src.mutators.GenTypeAwareMutation.Util import (
    type2num, get_all_subterms, local_compatible
)
from yinyang.src.parsing.Ast import Expr
from yinyang.src.parsing.Types import ALL


class GenTypeAwareMutation(Mutator):
    def __init__(self, formula, args, unique_expr):
        self.args = args
        self.formula = formula
        self.unique_expr = unique_expr
        self.operators = []
        self.parse_config_file()

    def parse_config_file(self):
        """
        Read the customizable configuration file.
        Customize the configuration file at yinyang/config/generalization.txt.
        Configuration file contains all the signatures of SMT-LIB operators
        and the signatures are used for operator choice during the mutation.
        """
        with open(self.args.config) as f:
            lines = f.readlines()

        for line in lines:
            line = line.strip("\n")  # remove linebreaks
            if ";" in line:  # ignore comments
                line = line[: line.index(";")]
            if line.strip() == "":  # skip empty lines
                continue
            line = line.strip()  # remove trailing whitespaces
            tokens = line.split(" ")

            if "par" not in tokens[0]:
                op_name, type_strings, attributes =\
                    handle_non_parametric_op(tokens)
                op = Operator(op_name, type_strings, attributes)
            else:
                op_name, type_strings, attributes, parameters =\
                    handle_parametric_op(tokens)
                op = Operator(op_name, type_strings, attributes, parameters)
            self.operators.append(op)

    def has_types(self, types):
        """
        types: list of types (possibly redundant)

        :returns: True if the seed formula supports the types
        """
        for t in set(types):
            if t == ALL:
                continue
            type_id = type2num[t]
            if len(self.unique_expr[type_id]) == 0:
                return False
        return True

    def get_candidate_ops(self, term):
        """
        term: term object

        An operator op is a candidate (op arg1 ... argk ret)
        (1) if its return type matches the type of t or has polymorphic return
            type (i.e. ret == ALL) and
        (2) if the seed formula has terms t1...tn such that tk.type = argk.type

        :returns: a list of candidate operators
        """
        candidate_ops = []
        for op in self.operators:
            if op.rtype != ALL and op.rtype != term.type:
                continue
            if self.has_types(op.arg_types):
                candidate_ops.append(op)
        return candidate_ops

    def get_replacee(self, term):
        """
        term: term object

        Choose randomly an operator from the list of candidate operators.

        When id is chosen for the operator, we pick an expression of a
        same type as input term. Else, we choose expression(s) according to
        the signature of the operator from the configuration file and generate
        a new expression.

        :returns: newly generated expression for substitution
        """
        candidate_ops = self.get_candidate_ops(term)

        if len(candidate_ops) == 0:
            return None

        op = random.choice(candidate_ops)
        args = []
        if op.name == "id":
            typ_id = type2num[term.type]
            if self.unique_expr[typ_id]:
                choices = [
                    termPrime
                    for termPrime in self.unique_expr[typ_id]
                    if termPrime != term and local_compatible(term, termPrime)
                ]

            if len(choices) == 0:
                return None

            return random.choice(choices)
        else:
            for t in op.arg_types:
                typ_id = type2num[t]
                choices = [tPrime for tPrime in self.unique_expr[typ_id]]

                if len(choices) == 0:
                    return None

                arg = random.choice(choices)
                args.append(arg)

            exp = Expr(op=op.name, subterms=args)
            exp.type = op.rtype

            return exp
        return None

    def mutate(self):
        """
        Perform a generative type-aware mutation.
        In case generator could not generate valid mutant, return false.

        :returns: mutant formula, and result of mutation
        """
        success = False
        self.av_expr, self.expr_type = get_all_subterms(self.formula)
        num_holes = len(self.av_expr)
        all_holes = self.av_expr
        for _ in range(num_holes):
            t1 = random.choice(all_holes)
            t2 = self.get_replacee(t1)
            if t2:
                success = True
                t1.substitute(t1, t2)
                break
            all_holes.remove(t1)
        return self.formula, success, False  # False = never skip seed



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import pathlib

from yinyang.src.base.Utils import in_list

try:
    sys.path.insert(1, os.getcwd() + "/.yinyang")
    from yinyang.config import crash_list, duplicate_list, ignore_list
except Exception as e:
    from yinyang.config.Config import crash_list, duplicate_list, ignore_list

from yinyang.src.core.Solver import SolverResult, SolverQueryResult


def in_crash_list(stdout, stderr):
    return in_list(stdout, stderr, crash_list)


def in_duplicate_list(stdout, stderr):
    return in_list(stdout, stderr, duplicate_list)


def in_ignore_list(stdout, stderr):
    return in_list(stdout, stderr, ignore_list)


def admissible_seed_size(seed, args):
    """
    Checks if seed size is below file_size_limit.
    :returns: True if that is the case and False otherwise.
    """
    seed_size_in_bytes = pathlib.Path(seed).stat().st_size
    if seed_size_in_bytes >= args.file_size_limit:
        return False
    return True


def grep_result(stdout):
    """
    Grep the result from the stdout of a solver.
    """
    result = SolverResult()
    for line in stdout.splitlines():
        if re.search("^unsat$", line, flags=re.MULTILINE):
            result.append(SolverQueryResult.UNSAT)
        elif re.search("^sat$", line, flags=re.MULTILINE):
            result.append(SolverQueryResult.SAT)
        elif re.search("^unknown$", line, flags=re.MULTILINE):
            result.append(SolverQueryResult.UNKNOWN)
    return result


def get_seeds(args, strategy):
    if strategy in ["opfuzz", "typefuzz"]:
        seeds = args.PATH_TO_SEEDS
    elif strategy == "yinyang":
        if len(args.PATH_TO_SEEDS) > 2:
            seeds = [(a, b)
                     for a in args.PATH_TO_SEEDS
                     for b in args.PATH_TO_SEEDS]
        elif len(args.PATH_TO_SEEDS) == 2:
            seeds = [(args.PATH_TO_SEEDS[0], args.PATH_TO_SEEDS[1])]
        else:
            assert False
    else:
        assert False

    return seeds


def init_oracle(args):
    """
    Initialize the oracle. For SemanticFusion the oracle is either sat or
    unsat. For TypeAwareOpMutation the oracle is unknown
    """
    if args.oracle == "unknown":
        return SolverResult(SolverQueryResult.UNKNOWN)
    elif args.oracle == "sat":
        return SolverResult(SolverQueryResult.SAT)
    elif args.oracle == "unsat":
        return SolverResult(SolverQueryResult.UNSAT)
    assert False



# Generated from SMTLIBv2.g4 by ANTLR 4.9.2
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .SMTLIBv2Parser import SMTLIBv2Parser
else:
    from SMTLIBv2Parser import SMTLIBv2Parser

# This class defines a complete listener for a parse tree produced by SMTLIBv2Parser.
class SMTLIBv2Listener(ParseTreeListener):

    # Enter a parse tree produced by SMTLIBv2Parser#start.
    def enterStart(self, ctx: SMTLIBv2Parser.StartContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#start.
    def exitStart(self, ctx: SMTLIBv2Parser.StartContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#response.
    def enterResponse(self, ctx: SMTLIBv2Parser.ResponseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#response.
    def exitResponse(self, ctx: SMTLIBv2Parser.ResponseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#generalReservedWord.
    def enterGeneralReservedWord(self, ctx: SMTLIBv2Parser.GeneralReservedWordContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#generalReservedWord.
    def exitGeneralReservedWord(self, ctx: SMTLIBv2Parser.GeneralReservedWordContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#simpleSymbol.
    def enterSimpleSymbol(self, ctx: SMTLIBv2Parser.SimpleSymbolContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#simpleSymbol.
    def exitSimpleSymbol(self, ctx: SMTLIBv2Parser.SimpleSymbolContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#quotedSymbol.
    def enterQuotedSymbol(self, ctx: SMTLIBv2Parser.QuotedSymbolContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#quotedSymbol.
    def exitQuotedSymbol(self, ctx: SMTLIBv2Parser.QuotedSymbolContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#predefSymbol.
    def enterPredefSymbol(self, ctx: SMTLIBv2Parser.PredefSymbolContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#predefSymbol.
    def exitPredefSymbol(self, ctx: SMTLIBv2Parser.PredefSymbolContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#predefKeyword.
    def enterPredefKeyword(self, ctx: SMTLIBv2Parser.PredefKeywordContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#predefKeyword.
    def exitPredefKeyword(self, ctx: SMTLIBv2Parser.PredefKeywordContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#symbol.
    def enterSymbol(self, ctx: SMTLIBv2Parser.SymbolContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#symbol.
    def exitSymbol(self, ctx: SMTLIBv2Parser.SymbolContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#numeral.
    def enterNumeral(self, ctx: SMTLIBv2Parser.NumeralContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#numeral.
    def exitNumeral(self, ctx: SMTLIBv2Parser.NumeralContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#decimal.
    def enterDecimal(self, ctx: SMTLIBv2Parser.DecimalContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#decimal.
    def exitDecimal(self, ctx: SMTLIBv2Parser.DecimalContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#hexadecimal.
    def enterHexadecimal(self, ctx: SMTLIBv2Parser.HexadecimalContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#hexadecimal.
    def exitHexadecimal(self, ctx: SMTLIBv2Parser.HexadecimalContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#binary.
    def enterBinary(self, ctx: SMTLIBv2Parser.BinaryContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#binary.
    def exitBinary(self, ctx: SMTLIBv2Parser.BinaryContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#string.
    def enterString(self, ctx: SMTLIBv2Parser.StringContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#string.
    def exitString(self, ctx: SMTLIBv2Parser.StringContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#reg_const.
    def enterReg_const(self, ctx: SMTLIBv2Parser.Reg_constContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#reg_const.
    def exitReg_const(self, ctx: SMTLIBv2Parser.Reg_constContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#keyword.
    def enterKeyword(self, ctx: SMTLIBv2Parser.KeywordContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#keyword.
    def exitKeyword(self, ctx: SMTLIBv2Parser.KeywordContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#spec_constant.
    def enterSpec_constant(self, ctx: SMTLIBv2Parser.Spec_constantContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#spec_constant.
    def exitSpec_constant(self, ctx: SMTLIBv2Parser.Spec_constantContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#s_expr.
    def enterS_expr(self, ctx: SMTLIBv2Parser.S_exprContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#s_expr.
    def exitS_expr(self, ctx: SMTLIBv2Parser.S_exprContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#index.
    def enterIndex(self, ctx: SMTLIBv2Parser.IndexContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#index.
    def exitIndex(self, ctx: SMTLIBv2Parser.IndexContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#identifier.
    def enterIdentifier(self, ctx: SMTLIBv2Parser.IdentifierContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#identifier.
    def exitIdentifier(self, ctx: SMTLIBv2Parser.IdentifierContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#attribute_value.
    def enterAttribute_value(self, ctx: SMTLIBv2Parser.Attribute_valueContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#attribute_value.
    def exitAttribute_value(self, ctx: SMTLIBv2Parser.Attribute_valueContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#attribute.
    def enterAttribute(self, ctx: SMTLIBv2Parser.AttributeContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#attribute.
    def exitAttribute(self, ctx: SMTLIBv2Parser.AttributeContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#sort.
    def enterSort(self, ctx: SMTLIBv2Parser.SortContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#sort.
    def exitSort(self, ctx: SMTLIBv2Parser.SortContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#qual_identifier.
    def enterQual_identifier(self, ctx: SMTLIBv2Parser.Qual_identifierContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#qual_identifier.
    def exitQual_identifier(self, ctx: SMTLIBv2Parser.Qual_identifierContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#var_binding.
    def enterVar_binding(self, ctx: SMTLIBv2Parser.Var_bindingContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#var_binding.
    def exitVar_binding(self, ctx: SMTLIBv2Parser.Var_bindingContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#sorted_var.
    def enterSorted_var(self, ctx: SMTLIBv2Parser.Sorted_varContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#sorted_var.
    def exitSorted_var(self, ctx: SMTLIBv2Parser.Sorted_varContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#pattern.
    def enterPattern(self, ctx: SMTLIBv2Parser.PatternContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#pattern.
    def exitPattern(self, ctx: SMTLIBv2Parser.PatternContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#match_case.
    def enterMatch_case(self, ctx: SMTLIBv2Parser.Match_caseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#match_case.
    def exitMatch_case(self, ctx: SMTLIBv2Parser.Match_caseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#term.
    def enterTerm(self, ctx: SMTLIBv2Parser.TermContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#term.
    def exitTerm(self, ctx: SMTLIBv2Parser.TermContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#sort_symbol_decl.
    def enterSort_symbol_decl(self, ctx: SMTLIBv2Parser.Sort_symbol_declContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#sort_symbol_decl.
    def exitSort_symbol_decl(self, ctx: SMTLIBv2Parser.Sort_symbol_declContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#meta_spec_constant.
    def enterMeta_spec_constant(self, ctx: SMTLIBv2Parser.Meta_spec_constantContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#meta_spec_constant.
    def exitMeta_spec_constant(self, ctx: SMTLIBv2Parser.Meta_spec_constantContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#fun_symbol_decl.
    def enterFun_symbol_decl(self, ctx: SMTLIBv2Parser.Fun_symbol_declContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#fun_symbol_decl.
    def exitFun_symbol_decl(self, ctx: SMTLIBv2Parser.Fun_symbol_declContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#par_fun_symbol_decl.
    def enterPar_fun_symbol_decl(self, ctx: SMTLIBv2Parser.Par_fun_symbol_declContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#par_fun_symbol_decl.
    def exitPar_fun_symbol_decl(self, ctx: SMTLIBv2Parser.Par_fun_symbol_declContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#theory_attribute.
    def enterTheory_attribute(self, ctx: SMTLIBv2Parser.Theory_attributeContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#theory_attribute.
    def exitTheory_attribute(self, ctx: SMTLIBv2Parser.Theory_attributeContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#theory_decl.
    def enterTheory_decl(self, ctx: SMTLIBv2Parser.Theory_declContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#theory_decl.
    def exitTheory_decl(self, ctx: SMTLIBv2Parser.Theory_declContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#logic_attribue.
    def enterLogic_attribue(self, ctx: SMTLIBv2Parser.Logic_attribueContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#logic_attribue.
    def exitLogic_attribue(self, ctx: SMTLIBv2Parser.Logic_attribueContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#logic.
    def enterLogic(self, ctx: SMTLIBv2Parser.LogicContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#logic.
    def exitLogic(self, ctx: SMTLIBv2Parser.LogicContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#sort_dec.
    def enterSort_dec(self, ctx: SMTLIBv2Parser.Sort_decContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#sort_dec.
    def exitSort_dec(self, ctx: SMTLIBv2Parser.Sort_decContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#selector_dec.
    def enterSelector_dec(self, ctx: SMTLIBv2Parser.Selector_decContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#selector_dec.
    def exitSelector_dec(self, ctx: SMTLIBv2Parser.Selector_decContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#constructor_dec.
    def enterConstructor_dec(self, ctx: SMTLIBv2Parser.Constructor_decContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#constructor_dec.
    def exitConstructor_dec(self, ctx: SMTLIBv2Parser.Constructor_decContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#datatype_dec.
    def enterDatatype_dec(self, ctx: SMTLIBv2Parser.Datatype_decContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#datatype_dec.
    def exitDatatype_dec(self, ctx: SMTLIBv2Parser.Datatype_decContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#function_dec.
    def enterFunction_dec(self, ctx: SMTLIBv2Parser.Function_decContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#function_dec.
    def exitFunction_dec(self, ctx: SMTLIBv2Parser.Function_decContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#function_def.
    def enterFunction_def(self, ctx: SMTLIBv2Parser.Function_defContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#function_def.
    def exitFunction_def(self, ctx: SMTLIBv2Parser.Function_defContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#prop_literal.
    def enterProp_literal(self, ctx: SMTLIBv2Parser.Prop_literalContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#prop_literal.
    def exitProp_literal(self, ctx: SMTLIBv2Parser.Prop_literalContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#script.
    def enterScript(self, ctx: SMTLIBv2Parser.ScriptContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#script.
    def exitScript(self, ctx: SMTLIBv2Parser.ScriptContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_assert.
    def enterCmd_assert(self, ctx: SMTLIBv2Parser.Cmd_assertContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_assert.
    def exitCmd_assert(self, ctx: SMTLIBv2Parser.Cmd_assertContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_assertSoft.
    def enterCmd_assertSoft(self, ctx: SMTLIBv2Parser.Cmd_assertSoftContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_assertSoft.
    def exitCmd_assertSoft(self, ctx: SMTLIBv2Parser.Cmd_assertSoftContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_simplify.
    def enterCmd_simplify(self, ctx: SMTLIBv2Parser.Cmd_simplifyContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_simplify.
    def exitCmd_simplify(self, ctx: SMTLIBv2Parser.Cmd_simplifyContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_checkSat.
    def enterCmd_checkSat(self, ctx: SMTLIBv2Parser.Cmd_checkSatContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_checkSat.
    def exitCmd_checkSat(self, ctx: SMTLIBv2Parser.Cmd_checkSatContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_checkSatAssuming.
    def enterCmd_checkSatAssuming(
        self, ctx: SMTLIBv2Parser.Cmd_checkSatAssumingContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_checkSatAssuming.
    def exitCmd_checkSatAssuming(self, ctx: SMTLIBv2Parser.Cmd_checkSatAssumingContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_checkSatUsing.
    def enterCmd_checkSatUsing(self, ctx: SMTLIBv2Parser.Cmd_checkSatUsingContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_checkSatUsing.
    def exitCmd_checkSatUsing(self, ctx: SMTLIBv2Parser.Cmd_checkSatUsingContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_minimize.
    def enterCmd_minimize(self, ctx: SMTLIBv2Parser.Cmd_minimizeContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_minimize.
    def exitCmd_minimize(self, ctx: SMTLIBv2Parser.Cmd_minimizeContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_maximize.
    def enterCmd_maximize(self, ctx: SMTLIBv2Parser.Cmd_maximizeContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_maximize.
    def exitCmd_maximize(self, ctx: SMTLIBv2Parser.Cmd_maximizeContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareConst.
    def enterCmd_declareConst(self, ctx: SMTLIBv2Parser.Cmd_declareConstContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareConst.
    def exitCmd_declareConst(self, ctx: SMTLIBv2Parser.Cmd_declareConstContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareDatatype.
    def enterCmd_declareDatatype(self, ctx: SMTLIBv2Parser.Cmd_declareDatatypeContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareDatatype.
    def exitCmd_declareDatatype(self, ctx: SMTLIBv2Parser.Cmd_declareDatatypeContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareCodatatype.
    def enterCmd_declareCodatatype(
        self, ctx: SMTLIBv2Parser.Cmd_declareCodatatypeContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareCodatatype.
    def exitCmd_declareCodatatype(
        self, ctx: SMTLIBv2Parser.Cmd_declareCodatatypeContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareDatatypes.
    def enterCmd_declareDatatypes(
        self, ctx: SMTLIBv2Parser.Cmd_declareDatatypesContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareDatatypes.
    def exitCmd_declareDatatypes(self, ctx: SMTLIBv2Parser.Cmd_declareDatatypesContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareCodatatypes.
    def enterCmd_declareCodatatypes(
        self, ctx: SMTLIBv2Parser.Cmd_declareCodatatypesContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareCodatatypes.
    def exitCmd_declareCodatatypes(
        self, ctx: SMTLIBv2Parser.Cmd_declareCodatatypesContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareFun.
    def enterCmd_declareFun(self, ctx: SMTLIBv2Parser.Cmd_declareFunContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareFun.
    def exitCmd_declareFun(self, ctx: SMTLIBv2Parser.Cmd_declareFunContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_declareSort.
    def enterCmd_declareSort(self, ctx: SMTLIBv2Parser.Cmd_declareSortContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_declareSort.
    def exitCmd_declareSort(self, ctx: SMTLIBv2Parser.Cmd_declareSortContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_define.
    def enterCmd_define(self, ctx: SMTLIBv2Parser.Cmd_defineContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_define.
    def exitCmd_define(self, ctx: SMTLIBv2Parser.Cmd_defineContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_defineFun.
    def enterCmd_defineFun(self, ctx: SMTLIBv2Parser.Cmd_defineFunContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_defineFun.
    def exitCmd_defineFun(self, ctx: SMTLIBv2Parser.Cmd_defineFunContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_defineConst.
    def enterCmd_defineConst(self, ctx: SMTLIBv2Parser.Cmd_defineConstContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_defineConst.
    def exitCmd_defineConst(self, ctx: SMTLIBv2Parser.Cmd_defineConstContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_defineFunRec.
    def enterCmd_defineFunRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunRecContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_defineFunRec.
    def exitCmd_defineFunRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunRecContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_defineFunsRec.
    def enterCmd_defineFunsRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunsRecContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_defineFunsRec.
    def exitCmd_defineFunsRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunsRecContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_defineSort.
    def enterCmd_defineSort(self, ctx: SMTLIBv2Parser.Cmd_defineSortContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_defineSort.
    def exitCmd_defineSort(self, ctx: SMTLIBv2Parser.Cmd_defineSortContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_display.
    def enterCmd_display(self, ctx: SMTLIBv2Parser.Cmd_displayContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_display.
    def exitCmd_display(self, ctx: SMTLIBv2Parser.Cmd_displayContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_echo.
    def enterCmd_echo(self, ctx: SMTLIBv2Parser.Cmd_echoContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_echo.
    def exitCmd_echo(self, ctx: SMTLIBv2Parser.Cmd_echoContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_eval.
    def enterCmd_eval(self, ctx: SMTLIBv2Parser.Cmd_evalContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_eval.
    def exitCmd_eval(self, ctx: SMTLIBv2Parser.Cmd_evalContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_exit.
    def enterCmd_exit(self, ctx: SMTLIBv2Parser.Cmd_exitContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_exit.
    def exitCmd_exit(self, ctx: SMTLIBv2Parser.Cmd_exitContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_GetObjectives.
    def enterCmd_GetObjectives(self, ctx: SMTLIBv2Parser.Cmd_GetObjectivesContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_GetObjectives.
    def exitCmd_GetObjectives(self, ctx: SMTLIBv2Parser.Cmd_GetObjectivesContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getAssertions.
    def enterCmd_getAssertions(self, ctx: SMTLIBv2Parser.Cmd_getAssertionsContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getAssertions.
    def exitCmd_getAssertions(self, ctx: SMTLIBv2Parser.Cmd_getAssertionsContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getAssignment.
    def enterCmd_getAssignment(self, ctx: SMTLIBv2Parser.Cmd_getAssignmentContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getAssignment.
    def exitCmd_getAssignment(self, ctx: SMTLIBv2Parser.Cmd_getAssignmentContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getInfo.
    def enterCmd_getInfo(self, ctx: SMTLIBv2Parser.Cmd_getInfoContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getInfo.
    def exitCmd_getInfo(self, ctx: SMTLIBv2Parser.Cmd_getInfoContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getModel.
    def enterCmd_getModel(self, ctx: SMTLIBv2Parser.Cmd_getModelContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getModel.
    def exitCmd_getModel(self, ctx: SMTLIBv2Parser.Cmd_getModelContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_blockModel.
    def enterCmd_blockModel(self, ctx: SMTLIBv2Parser.Cmd_blockModelContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_blockModel.
    def exitCmd_blockModel(self, ctx: SMTLIBv2Parser.Cmd_blockModelContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getOption.
    def enterCmd_getOption(self, ctx: SMTLIBv2Parser.Cmd_getOptionContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getOption.
    def exitCmd_getOption(self, ctx: SMTLIBv2Parser.Cmd_getOptionContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getProof.
    def enterCmd_getProof(self, ctx: SMTLIBv2Parser.Cmd_getProofContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getProof.
    def exitCmd_getProof(self, ctx: SMTLIBv2Parser.Cmd_getProofContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getUnsatAssumptions.
    def enterCmd_getUnsatAssumptions(
        self, ctx: SMTLIBv2Parser.Cmd_getUnsatAssumptionsContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getUnsatAssumptions.
    def exitCmd_getUnsatAssumptions(
        self, ctx: SMTLIBv2Parser.Cmd_getUnsatAssumptionsContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_labels.
    def enterCmd_labels(self, ctx: SMTLIBv2Parser.Cmd_labelsContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_labels.
    def exitCmd_labels(self, ctx: SMTLIBv2Parser.Cmd_labelsContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getUnsatCore.
    def enterCmd_getUnsatCore(self, ctx: SMTLIBv2Parser.Cmd_getUnsatCoreContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getUnsatCore.
    def exitCmd_getUnsatCore(self, ctx: SMTLIBv2Parser.Cmd_getUnsatCoreContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_getValue.
    def enterCmd_getValue(self, ctx: SMTLIBv2Parser.Cmd_getValueContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_getValue.
    def exitCmd_getValue(self, ctx: SMTLIBv2Parser.Cmd_getValueContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_pop.
    def enterCmd_pop(self, ctx: SMTLIBv2Parser.Cmd_popContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_pop.
    def exitCmd_pop(self, ctx: SMTLIBv2Parser.Cmd_popContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_poly_factor.
    def enterCmd_poly_factor(self, ctx: SMTLIBv2Parser.Cmd_poly_factorContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_poly_factor.
    def exitCmd_poly_factor(self, ctx: SMTLIBv2Parser.Cmd_poly_factorContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_push.
    def enterCmd_push(self, ctx: SMTLIBv2Parser.Cmd_pushContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_push.
    def exitCmd_push(self, ctx: SMTLIBv2Parser.Cmd_pushContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_reset.
    def enterCmd_reset(self, ctx: SMTLIBv2Parser.Cmd_resetContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_reset.
    def exitCmd_reset(self, ctx: SMTLIBv2Parser.Cmd_resetContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_resetAssertions.
    def enterCmd_resetAssertions(self, ctx: SMTLIBv2Parser.Cmd_resetAssertionsContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_resetAssertions.
    def exitCmd_resetAssertions(self, ctx: SMTLIBv2Parser.Cmd_resetAssertionsContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_setInfo.
    def enterCmd_setInfo(self, ctx: SMTLIBv2Parser.Cmd_setInfoContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_setInfo.
    def exitCmd_setInfo(self, ctx: SMTLIBv2Parser.Cmd_setInfoContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_setLogic.
    def enterCmd_setLogic(self, ctx: SMTLIBv2Parser.Cmd_setLogicContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_setLogic.
    def exitCmd_setLogic(self, ctx: SMTLIBv2Parser.Cmd_setLogicContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#cmd_setOption.
    def enterCmd_setOption(self, ctx: SMTLIBv2Parser.Cmd_setOptionContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#cmd_setOption.
    def exitCmd_setOption(self, ctx: SMTLIBv2Parser.Cmd_setOptionContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#tac_then.
    def enterTac_then(self, ctx: SMTLIBv2Parser.Tac_thenContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#tac_then.
    def exitTac_then(self, ctx: SMTLIBv2Parser.Tac_thenContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#tac_and_then.
    def enterTac_and_then(self, ctx: SMTLIBv2Parser.Tac_and_thenContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#tac_and_then.
    def exitTac_and_then(self, ctx: SMTLIBv2Parser.Tac_and_thenContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#par_then.
    def enterPar_then(self, ctx: SMTLIBv2Parser.Par_thenContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#par_then.
    def exitPar_then(self, ctx: SMTLIBv2Parser.Par_thenContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#or_else.
    def enterOr_else(self, ctx: SMTLIBv2Parser.Or_elseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#or_else.
    def exitOr_else(self, ctx: SMTLIBv2Parser.Or_elseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#par_or_else.
    def enterPar_or_else(self, ctx: SMTLIBv2Parser.Par_or_elseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#par_or_else.
    def exitPar_or_else(self, ctx: SMTLIBv2Parser.Par_or_elseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#par_or.
    def enterPar_or(self, ctx: SMTLIBv2Parser.Par_orContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#par_or.
    def exitPar_or(self, ctx: SMTLIBv2Parser.Par_orContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#tryFor.
    def enterTryFor(self, ctx: SMTLIBv2Parser.TryForContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#tryFor.
    def exitTryFor(self, ctx: SMTLIBv2Parser.TryForContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#usingParams.
    def enterUsingParams(self, ctx: SMTLIBv2Parser.UsingParamsContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#usingParams.
    def exitUsingParams(self, ctx: SMTLIBv2Parser.UsingParamsContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#tactical.
    def enterTactical(self, ctx: SMTLIBv2Parser.TacticalContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#tactical.
    def exitTactical(self, ctx: SMTLIBv2Parser.TacticalContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#command.
    def enterCommand(self, ctx: SMTLIBv2Parser.CommandContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#command.
    def exitCommand(self, ctx: SMTLIBv2Parser.CommandContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#b_value.
    def enterB_value(self, ctx: SMTLIBv2Parser.B_valueContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#b_value.
    def exitB_value(self, ctx: SMTLIBv2Parser.B_valueContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#option.
    def enterOption(self, ctx: SMTLIBv2Parser.OptionContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#option.
    def exitOption(self, ctx: SMTLIBv2Parser.OptionContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#info_flag.
    def enterInfo_flag(self, ctx: SMTLIBv2Parser.Info_flagContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#info_flag.
    def exitInfo_flag(self, ctx: SMTLIBv2Parser.Info_flagContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#error_behaviour.
    def enterError_behaviour(self, ctx: SMTLIBv2Parser.Error_behaviourContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#error_behaviour.
    def exitError_behaviour(self, ctx: SMTLIBv2Parser.Error_behaviourContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#reason_unknown.
    def enterReason_unknown(self, ctx: SMTLIBv2Parser.Reason_unknownContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#reason_unknown.
    def exitReason_unknown(self, ctx: SMTLIBv2Parser.Reason_unknownContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#model_response.
    def enterModel_response(self, ctx: SMTLIBv2Parser.Model_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#model_response.
    def exitModel_response(self, ctx: SMTLIBv2Parser.Model_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#info_response.
    def enterInfo_response(self, ctx: SMTLIBv2Parser.Info_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#info_response.
    def exitInfo_response(self, ctx: SMTLIBv2Parser.Info_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#valuation_pair.
    def enterValuation_pair(self, ctx: SMTLIBv2Parser.Valuation_pairContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#valuation_pair.
    def exitValuation_pair(self, ctx: SMTLIBv2Parser.Valuation_pairContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#t_valuation_pair.
    def enterT_valuation_pair(self, ctx: SMTLIBv2Parser.T_valuation_pairContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#t_valuation_pair.
    def exitT_valuation_pair(self, ctx: SMTLIBv2Parser.T_valuation_pairContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#check_sat_response.
    def enterCheck_sat_response(self, ctx: SMTLIBv2Parser.Check_sat_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#check_sat_response.
    def exitCheck_sat_response(self, ctx: SMTLIBv2Parser.Check_sat_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#echo_response.
    def enterEcho_response(self, ctx: SMTLIBv2Parser.Echo_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#echo_response.
    def exitEcho_response(self, ctx: SMTLIBv2Parser.Echo_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_assertions_response.
    def enterGet_assertions_response(
        self, ctx: SMTLIBv2Parser.Get_assertions_responseContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_assertions_response.
    def exitGet_assertions_response(
        self, ctx: SMTLIBv2Parser.Get_assertions_responseContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_assignment_response.
    def enterGet_assignment_response(
        self, ctx: SMTLIBv2Parser.Get_assignment_responseContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_assignment_response.
    def exitGet_assignment_response(
        self, ctx: SMTLIBv2Parser.Get_assignment_responseContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_info_response.
    def enterGet_info_response(self, ctx: SMTLIBv2Parser.Get_info_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_info_response.
    def exitGet_info_response(self, ctx: SMTLIBv2Parser.Get_info_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_model_response.
    def enterGet_model_response(self, ctx: SMTLIBv2Parser.Get_model_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_model_response.
    def exitGet_model_response(self, ctx: SMTLIBv2Parser.Get_model_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_option_response.
    def enterGet_option_response(self, ctx: SMTLIBv2Parser.Get_option_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_option_response.
    def exitGet_option_response(self, ctx: SMTLIBv2Parser.Get_option_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_proof_response.
    def enterGet_proof_response(self, ctx: SMTLIBv2Parser.Get_proof_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_proof_response.
    def exitGet_proof_response(self, ctx: SMTLIBv2Parser.Get_proof_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_unsat_assump_response.
    def enterGet_unsat_assump_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_assump_responseContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_unsat_assump_response.
    def exitGet_unsat_assump_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_assump_responseContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_unsat_core_response.
    def enterGet_unsat_core_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_core_responseContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_unsat_core_response.
    def exitGet_unsat_core_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_core_responseContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#get_value_response.
    def enterGet_value_response(self, ctx: SMTLIBv2Parser.Get_value_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#get_value_response.
    def exitGet_value_response(self, ctx: SMTLIBv2Parser.Get_value_responseContext):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#specific_success_response.
    def enterSpecific_success_response(
        self, ctx: SMTLIBv2Parser.Specific_success_responseContext
    ):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#specific_success_response.
    def exitSpecific_success_response(
        self, ctx: SMTLIBv2Parser.Specific_success_responseContext
    ):
        pass

    # Enter a parse tree produced by SMTLIBv2Parser#general_response.
    def enterGeneral_response(self, ctx: SMTLIBv2Parser.General_responseContext):
        pass

    # Exit a parse tree produced by SMTLIBv2Parser#general_response.
    def exitGeneral_response(self, ctx: SMTLIBv2Parser.General_responseContext):
        pass


del SMTLIBv2Parser



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

solvers = []

crash_list = [
    "Exception",
    "lang.AssertionError",
    "lang.Error",
    "runtime error",
    "LEAKED",
    "Leaked",
    "Segmentation fault",
    "segmentation fault",
    "segfault",
    "ASSERTION",
    "Assertion",
    "Fatal failure",
    "Internal error detected",
    "an invalid model was generated",
    "Failed to verify",
    "failed to verify",
    "ERROR: AddressSanitizer:",
    "invalid expression",
    "Aborted"
]

duplicate_list = [
    "src/smt/smt_mock.cpp:1489"
]

ignore_list = [
    "(error ",
    "unsupported",
    "unexpected char",
    "failed to open file",
    "Expected result sat but got unsat",
    "Expected result unsat but got sat",
    "Parse Error",
    "Cannot get model",
    "Symbol 'str.to-re' not declared as a variable",
    "Symbol 'str.to.re' not declared as a variable",
    "Unimplemented code encountered",
]



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import copy

from yinyang.src.mutators.Mutator import Mutator
from yinyang.src.mutators.SemanticFusion.VariableFusion import (
    fill_template,
    inv_x,
    inv_y,
    fusion_contraints,
    add_fusion_constraints,
    add_var_decls,
)
from yinyang.src.mutators.SemanticFusion.Util import (
    random_var_triplets,
    disjunction,
    conjunction,
)
from yinyang.src.base.Exitcodes import ERR_USAGE
from yinyang.src.parsing.Parse import parse_str
from yinyang.src.parsing.Ast import DeclareFun


class SemanticFusion(Mutator):
    def __init__(self, formula1, formula2, args):
        self.formula1 = formula1
        self.formula2 = formula2
        self.args = args
        self.config = self.args.config
        self.oracle = self.args.oracle
        self.templates = {}
        self._parse_mrs()

        if not self.oracle:
            print("error: No oracle {sat,unsat} specified")
            exit(ERR_USAGE)

    def _parse_mrs(self):
        with open(self.config) as f:
            lines = f.readlines()
        started = False
        curr = []
        _mrs = []

        for line in lines:
            if ";" in line:
                continue
            if not line.strip():
                continue
            if "begin" in line:
                started = True
                continue
            if "end" in line:
                started = False
                _mrs.append("\n".join(curr))
                curr = []
                continue
            if started:
                curr.append(line)

        for i, mr in enumerate(_mrs):
            template, _ = parse_str(mr)
            sort = template.commands[0].sort

            if sort not in self.templates:
                self.templates[sort] = [template]
            else:
                self.templates[sort].append(template)

    def fuse(self, formula1, formula2, triplets):

        fusion_vars = []
        fusion_constr = []
        for triplet in triplets:
            x, y, template, var_type =\
                triplet[0], triplet[1], triplet[2], triplet[3]
            z = DeclareFun(x + "_" + y + "_fused", "", var_type)
            fusion_vars.append(z)
            template = fill_template(x, y, template, var_type)
            fusion_constr += fusion_contraints(template, var_type)

            occs_x = [occ for occ in formula1.free_var_occs if occ.name == x]
            occs_y = [occ for occ in formula2.free_var_occs if occ.name == y]

            k = random.randint(0, len(occs_x))
            occs_x = random.sample(occs_x, k)
            k = random.randint(0, len(occs_y))
            occs_y = random.sample(occs_y, k)

            for occ in occs_x:
                occ.substitute(occ, inv_x(template))
            for occ in occs_y:
                occ.substitute(occ, inv_y(template))

        if self.oracle == "unsat":
            formula = disjunction(formula1, formula2)
            add_fusion_constraints(formula, fusion_constr)
        else:
            formula = conjunction(formula1, formula2)
        add_var_decls(formula, fusion_vars)

        return formula

    def mutate(self):
        skip_seed = False
        if self.formula1.free_var_occs == [] and\
           self.formula2.free_var_occs == []:
            skip_seed = True

        formula1, formula2 =\
            copy.deepcopy(self.formula1), copy.deepcopy(self.formula2)
        formula1.prefix_vars("scr1_")
        formula2.prefix_vars("scr2_")

        triplets = random_var_triplets(
            formula1.global_vars, formula2.global_vars, self.templates
        )
        fused = self.fuse(formula1, formula2, triplets)
        return fused, True, skip_seed



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yinyang.src.parsing.Types import sort2type


class Operator:
    def __init__(self, name, types, attributes, parameters=[]):
        self.name = name
        self.arg_types = [sort2type(t) for t in types[:-1]]
        self.rtype = sort2type(types[-1])
        self.attributes = attributes

    def __str__(self):
        s = self.name + ":" + ",".join(self.arg_types) + " -> " + self.rtype
        if self.attributes != []:
            s += " " + " ".join(self.attributes)
        return s

    def __repr__(self):
        return self.__str__()


def handle_non_parametric_op(tokens):
    op_name, type_strings, attributes = "", [], []
    for idx, tok in enumerate(tokens):
        if tok == "":
            continue  # ignore empty token
        if idx == 0:  # first token is the operator name
            op_name = tok[1:]
        elif ":" != tok[0]:
            type_strings.append(tok.strip(")"))
        else:
            attributes.append(tok.strip(")"))
    return op_name, type_strings, attributes


def handle_parametric_op(tokens):
    op_name, type_strings, attributes, parameters = "", [], [], []
    for idx, tok in enumerate(tokens):
        if idx == 0:
            continue
        if idx == 1:
            parameters.append(tok.strip("(").strip(")"))
            continue
        if idx == 2:
            op_name = tok[1:]
        elif ":" != tok[0]:
            type_strings.append(tok.strip(")"))
        else:
            attributes.append(tok.strip(")"))
    return op_name, type_strings, attributes, parameters



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
from pathlib import Path
from yinyang.src.base.Exitcodes import ERR_USAGE, ERR_EXHAUSTED_DISK

path = Path(__file__)
rootpath = str(path.parent.absolute().parent)

try:
    sys.path.insert(1, os.getcwd() + "/.yinyang")
    from yinyang.config import solvers
except Exception as e:
    from yinyang.config.Config import solvers


def check_solver_clis():
    if args.SOLVER_CLIS == "":
        if len(solvers) == 0:
            print("error: no solver specified", flush=True)
            exit(ERR_USAGE)
        args.SOLVER_CLIS = solvers
    else:
        args.SOLVER_CLIS = args.SOLVER_CLIS.split(";") + solvers


def check_timeout():
    if args.timeout <= 0:
        print("error: timeout should not be a negative number or zero",
              flush=True)
        exit(ERR_USAGE)


def check_iterations():
    if args.iterations <= 0:
        print("error: iterations should not be a negative number zero",
              flush=True)
        exit(ERR_USAGE)


def create_bug_folder():
    if not os.path.isdir(args.bugsfolder):
        try:
            os.mkdir(args.bugsfolder)
        except Exception:
            print("error: bug folder cannot be created", flush=True)
            exit(ERR_EXHAUSTED_DISK)


def create_log_folder():
    if not os.path.isdir(args.logfolder):
        try:
            os.mkdir(args.logfolder)
        except Exception:
            print("error: log folder cannot be created", flush=True)
            exit(ERR_EXHAUSTED_DISK)


def create_scratch_folder():
    if not os.path.isdir(args.scratchfolder):
        try:
            os.mkdir(args.scratchfolder)
        except Exception:
            print("error: scratch folder cannot be created", flush=True)
            exit(ERR_EXHAUSTED_DISK)


def get_seeds():
    temp_seeds = []
    for path in args.PATH_TO_SEEDS:
        if not os.path.exists(path):
            print('error: folder/file "%s" does not exist' % (path),
                  flush=True)
            exit(ERR_USAGE)
        if os.path.isfile(path):
            temp_seeds.append(path)
        elif os.path.isdir(path):
            for subdir, dirs, files in os.walk(path):
                for filename in files:
                    filepath = subdir + os.sep + filename
                    if filepath.endswith(".smt2"):
                        temp_seeds.append(filepath)
        else:
            print("error: %s is neither a file nor a directory", flush=True)
            exit(ERR_USAGE)

    args.PATH_TO_SEEDS = temp_seeds


def check_opfuzz():
    if len(args.PATH_TO_SEEDS) < 1:
        print("error: please provide at least one seed", flush=True)
        exit(ERR_USAGE)

    if len(args.SOLVER_CLIS) < 2:
        print("error: please provide at least two SMT solvers", flush=True)
        exit(ERR_USAGE)


def check_fusion():
    if len(args.PATH_TO_SEEDS) < 2:
        print(
            "error: please provide at least two seeds for the fusion strategy",
            flush=True,
        )
        exit(ERR_USAGE)

    if len(args.SOLVER_CLIS) < 1:
        print("error: please provide at one SMT solvers", flush=True)
        exit(ERR_USAGE)


def run_checks(parser, strategy):
    global args
    args = parser.parse_args()
    if not args.PATH_TO_SEEDS:
        parser.error("no seed-file/seed folder specified")

    check_solver_clis()
    check_timeout()
    check_iterations()
    create_bug_folder()
    create_log_folder()
    create_scratch_folder()
    get_seeds()
    if strategy in ["opfuzz", "typefuzz"]:
        check_opfuzz()
    else:
        check_fusion()
    return args



# MIT License
#
# Copyright (c) [2020 - 2020] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This file is about option fuzzer support.
# OptionGenerator reads a options setting file to generate a random option
# configuration for the corresponding solvers.
# The format of the options setting file is as follow:

# ```
# <solver keywords>
# <option name> [type|value]*
# <option name> [type|value]*
# ###
# <solver keywords>
# <option name> [type|value]*
# <option name> [type|value]*
# ###
# <solver keywords>
# <option name> [type|value]*
# <option name> [type|value]*
# ```

# `<solver keywords>`: Keywords for the solver command-line interfaces.
#                      If the keywords are found in the command-line interfaces
#                      yinyang will generate a corresponding random option
#                      setting.

# `<option name>`: Name of the option item.

# `[type|value]*`: Type or value of the option.
#                  The type can be either `bool` or `int`.
#                  The value can be either `true`,`false`, or an integer.
#                  In default, i.e., leaving this position empty, the option is
#                  assigned to `bool` type. Optfuzz will generate a random
#                  value according to the type of the option.

# `###`: Mark for splitting the option setting blocks.
#        The options in different blocks are independent.


import random

from enum import Enum


class OptionType(Enum):
    BOOL = 0
    INT = 1


class Option:
    def __init__(self, name, type, min=0, max=1000):
        self.name = name
        self.type = type
        self.min = min
        self.max = max


class ConfigBlock:
    def __init__(self):
        self.cmd = ""
        self.optionList = []


class OptionGenerator:
    def __init__(self, fn):
        with open(fn, "r") as reader:
            text = reader.read()
        self.configBlocks = []
        self.parse(text)

    def generate(self, cli):
        ret = ""
        for block in self.configBlocks:
            if block.cmd in cli:
                for option in block.optionList:
                    if option.type == OptionType.BOOL:
                        if option.max == option.min:
                            if option.max == 0:
                                boolean = "false"
                            else:
                                boolean = "true"
                        else:
                            if bool(random.getrandbits(1)):
                                boolean = "true"
                            else:
                                boolean = "false"
                        ret += "(set-option :%s %s)\n" % (option.name, boolean)
                    elif option.type == OptionType.INT:
                        ret += "(set-option :%s %s)\n" % (
                            option.name,
                            str(random.randrange(option.min, option.max + 1)),
                        )
        return ret

    def parse(self, text):
        configBlocks = text.split("###")
        for item in configBlocks:
            configBlock = ConfigBlock()
            configs = item.split("\n")
            for config in configs:
                if config != "":
                    if configBlock.cmd == "":
                        configBlock.cmd = config
                        continue
                    config = config.strip(" ").split(" ")
                    if len(config) == 1:
                        configBlock.optionList.append(
                            Option(config[0], OptionType.BOOL)
                        )
                    elif len(config) == 2:
                        if config[1] == "bool":
                            configBlock.optionList.append(
                                Option(config[0], OptionType.BOOL)
                            )
                        elif config[1] == "true":
                            configBlock.optionList.append(
                                Option(config[0], OptionType.BOOL, 1, 1)
                            )
                        elif config[1] == "false":
                            configBlock.optionList.append(
                                Option(config[0], OptionType.BOOL, 0, 0)
                            )
                        elif config[1] == "int":
                            configBlock.optionList.append(
                                Option(config[0], OptionType.INT)
                            )
                        elif config[1].isdigit():
                            configBlock.optionList.append(
                                Option(
                                    config[0],
                                    OptionType.INT,
                                    int(config[1]),
                                    int(config[1]),
                                )
                            )
                        elif "-" in config[1]:
                            bound = config[1].split("-")
                            if (
                                len(bound) == 2
                                and bound[0].isdigit()
                                and bound[1].isdigit()
                                and int(bound[0]) <= int(bound[1])
                            ):
                                configBlock.optionList.append(
                                    Option(
                                        config[0],
                                        OptionType.INT,
                                        int(bound[0]),
                                        int(bound[1]),
                                    )
                                )
                            else:
                                exit(
                                    "error: The range of the config is not\
                                     valid (e.g. 1-10):"
                                    + config[1]
                                )
                        else:
                            exit(
                                "error: The argument setting of the config is\
                                not valid (e.g. proof true): "
                                + config[1]
                            )
                    else:
                        exit(
                            "error: The argument setting of the config is not\
                             valid (e.g. proof true):"
                            + config[1]
                        )
            self.configBlocks.append(configBlock)



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import copy


class Script:
    def __init__(self, commands, global_vars):
        self.commands = commands
        self.vars, self.types = self._decl_commands()
        self.global_vars = global_vars
        self.free_var_occs = []
        self.op_occs = []
        self.assert_cmd = []

        for cmd in self.commands:
            if isinstance(cmd, Assert):
                globs_ = copy.deepcopy(self.global_vars)
                self._get_free_var_occs(cmd.term, self.global_vars)
                self.global_vars = globs_
                self._get_op_occs(cmd.term)
                self.assert_cmd.append(cmd)

    def _get_op_occs(self, e):
        if isinstance(e, str):
            return
        if e.is_const:
            return
        if e.label:
            return
        if e.is_var:
            return

        self.op_occs.append(e)
        for sub in e.subterms:
            self._get_op_occs(sub)

    def _get_free_var_occs(self, e, global_vars):
        if isinstance(e, str):
            return
        if e.is_const:
            return
        if e.label:
            return
        if e.quantifier:
            for var in list(global_vars):
                for quantified_var in e.quantified_vars:
                    if var == quantified_var[0]:
                        global_vars.pop(var)

        if e.var_binders:
            for var in list(global_vars):
                for let_var in e.var_binders:
                    if var == let_var:
                        global_vars.pop(var)
            for let_term in e.let_terms:
                self._get_free_var_occs(let_term, global_vars)

        if e.is_var:
            if e.name in global_vars:
                self.free_var_occs.append(e)
            return

        for sub in e.subterms:
            self._get_free_var_occs(sub, global_vars)

    def _decl_commands(self):
        vars, types = [], {}
        for cmd in self.commands:
            if isinstance(cmd, DeclareConst):
                vars.append(Var(cmd.symbol, cmd.sort))
                types[cmd.symbol] = cmd.sort
            if isinstance(cmd, DeclareFun):
                if cmd.input_sort != "":
                    vars.append(Var(cmd.symbol, cmd.input_sort))
                    types[cmd.symbol] = cmd.input_sort
        return vars, types

    def _prefix_free_vars(self, prefix, e):
        if isinstance(e, str):
            return
        if e.is_const:
            return
        if e.is_var and e.type:
            if e in self.free_var_occs:
                e.name = prefix + e.name
            return

        if e.var_binders:
            for i, var in enumerate(e.var_binders):
                self._prefix_free_vars(prefix, e.let_terms[i])

        for s in e.subterms:
            self._prefix_free_vars(prefix, s)

    def prefix_vars(self, prefix):
        """
        Add a shared prefix to all variables
        :prefix: str
        """
        for cmd in self.commands:
            if isinstance(cmd, DeclareConst):
                cmd.symbol = prefix + cmd.symbol
            if isinstance(cmd, DeclareFun):
                if cmd.input_sort == "":
                    cmd.symbol = prefix + cmd.symbol
            if isinstance(cmd, Assert):
                self._prefix_free_vars(prefix, cmd.term)
        new_global_vars = {}
        for global_var in self.global_vars:
            new_global_vars[prefix + global_var] = self.global_vars[global_var]
        self.global_vars = new_global_vars
        self.vars, self.types = self._decl_commands()

    def merge_asserts(self):
        """
        Merge all assert blocks (possibly separated by exit, reset,
        reset-assertions statement) into a single assert by conjunction.
        """
        terms = []
        for cmd in self.commands:
            if isinstance(cmd, Assert):
                terms.append(cmd.term)
            if isinstance(cmd, SMTLIBCommand):
                if cmd.cmd_str == "(exit)":
                    break
                if cmd.cmd_str == "(reset)":
                    terms = []
                if cmd.cmd_str == "(reset-assertions)":
                    terms = []
        conjunction = Assert(Term(op="and", subterms=terms))
        new_cmds, first_found = [], False
        for cmd in self.commands:
            if not first_found and isinstance(cmd, Assert):
                new_cmds.append(conjunction)
                first_found = True
            if isinstance(cmd, Assert):
                continue
            if isinstance(cmd, SMTLIBCommand):
                if cmd.cmd_str == "(exit)":
                    break
                if cmd.cmd_str == "(reset-assertions)":
                    continue
                if cmd.cmd_str == "(reset)":
                    new_cmds, first_found = [], False
                    continue
            new_cmds.append(cmd)
        self.commands = new_cmds

    def __str__(self):
        s = ""
        for i, c in enumerate(self.commands):
            if i != len(self.commands) - 1:
                s += c.__str__() + "\n"
            else:
                s += c.__str__()
        return s


class Commands:
    def __init__(self):
        self.free_vars = []


class DeclareConst(Commands):
    def __init__(self, symbol, sort):
        self.symbol = symbol
        self.sort = sort

    def __str__(self):
        return "(declare-const "\
               + self.symbol + " "\
               + self.sort.__str__() + ")"


class DeclareFun:
    def __init__(self, symbol, input_sort, output_sort):
        self.symbol = symbol
        self.input_sort = input_sort
        self.output_sort = output_sort

    def __str__(self):
        return (
            "(declare-fun "
            + self.symbol
            + " ("
            + self.input_sort
            + ") "
            + self.output_sort
            + ")"
        )


class Assert:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(assert " + self.term.__str__() + ")"


class AssertSoft:
    def __init__(self, term, attr):
        self.term = term
        self.attr = attr

    def __str__(self):
        for a in self.attr:
            attr_s = " " + a[0] + " " + a[1]
        return "(assert-soft " + self.term.__str__() + attr_s + ")"


class Comment:
    def __init__(self, txt):
        self.txt = txt

    def __str__(self):
        return "; " + self.txt


class Define:
    def __init__(self, symbol, term):
        self.term = term
        self.symbol = symbol

    def __str__(self):
        return "(define " + self.symbol + " " + self.term.__str__() + ")"


class DefineConst:
    def __init__(self, symbol, sort, term):
        self.symbol = symbol
        self.sort = sort
        self.term = term

    def __str__(self):
        return (
            "(define-const "
            + self.symbol
            + " "
            + self.sort
            + " "
            + self.term.__str__()
            + ")"
        )


class DefineFun:
    def __init__(self, symbol, sorted_vars, sort, term):
        self.symbol = symbol
        self.sorted_vars = sorted_vars
        self.sort = sort
        self.term = term

    def __str__(self):
        return (
            "(define-fun "
            + self.symbol
            + " ("
            + self.sorted_vars
            + ") "
            + self.sort
            + " "
            + self.term.__str__()
            + ")"
        )


class DefineFunRec:
    def __init__(self, symbol, sorted_vars, sort, term):
        self.symbol = symbol
        self.sorted_vars = sorted_vars
        self.sort = sort
        self.term = term

    def __str__(self):
        s = ""
        if len(self.sorted_vars) > 0:
            s = self.sorted_vars[0]
            for var in self.sorted_vars[1:]:
                s += " " + var
        return (
            "(define-fun-rec "
            + self.symbol
            + " ("
            + s
            + ") "
            + self.sort
            + " "
            + self.term.__str__()
            + ")"
        )


class FunDecl:
    def __init__(self, symbol, sorted_vars, sort):
        self.symbol = symbol
        self.sorted_vars = sorted_vars
        self.sort = sort

    def __str__(self):
        s = self.sorted_vars[0]
        for svar in self.sorted_vars[1:]:
            s += " " + svar
        return "(" + self.symbol + " (" + s + ") " + self.sort + ")"


class DefineFunsRec:
    def __init__(self, fun_decls, terms):
        self.fun_decls = fun_decls
        self.terms = terms

    def __str__(self):
        s = "(define-funs-rec (" + self.fun_decls[0].__str__()
        if len(self.fun_decls) > 1:
            for decl in self.fun_decls[1:]:
                s += " " + decl.__str__()
            s += ") (" + self.terms[0].__str__()
        if len(self.terms) > 1:
            for term in self.terms[1:]:
                s += " " + term.__str__()
            s += ")"
        return s + ")"


class Simplify:
    def __init__(self, term, attr):
        self.term = term
        self.attr = attr

    def __str__(self):
        attr_s = ""
        for a in self.attr:
            attr_s = " " + a[0] + " " + a[1]
        return "(simplify " + self.term.__str__() + attr_s + ")"


class Minimize:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(minimize " + self.term.__str__() + ")"


class Maximize:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(maximize " + self.term.__str__() + ")"


class Display:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(display" + self.term.__str__() + ")"


class Eval:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(eval" + self.term.__str__() + ")"


class PolyFactor:
    def __init__(self, term):
        self.term = term

    def __str__(self):
        return "(poly/factor" + self.term.__str__() + ")"


class CheckSat:
    def __init__(self, terms=None):
        self.terms = terms

    def __str__(self):
        t_str = ""
        if self.terms:
            t_str = ""
            for t in self.terms:
                t_str += " " + t.__str__()
            return "(check-sat" + t_str + ")"
        return "(check-sat)"


class CheckSatAssuming:
    def __init__(self, terms):
        self.terms = terms

    def __str__(self):
        s_term = self.terms[0].__str__()
        for t in self.terms[1:]:
            s_term += " " + t.__str__()
        return "(check-sat-assuming (" + s_term + "))"


class GetValue:
    def __init__(self, terms):
        self.terms = terms

    def __str__(self):
        t_str = ""
        for t in self.terms:
            t_str += t.__str__()
        return "(get-value (" + t_str + "))"


class Push:
    def __init__(self, terms=None):
        self.terms = terms

    def __str__(self):
        t_str = ""
        if self.terms:
            t_str = ""
            for t in self.terms:
                t_str += " " + t.__str__()
            return "(push" + t_str + ")"
        return "(push)"


class Pop:
    def __init__(self, terms=None):
        self.terms = terms

    def __str__(self):
        t_str = ""
        if self.terms:
            t_str = ""
            for t in self.terms:
                t_str += " " + t.__str__()
            return "(pop" + t_str + ")"
        return "(pop)"


class SMTLIBCommand:
    def __init__(self, cmd_str):
        self.cmd_str = cmd_str

    def __str__(self):
        return self.cmd_str

    def __eq__(self, other):
        if other.cmd_str == self.cmd_str:
            return True
        return False

    def __hash__(self):
        return self.cmd_str.__hash__()


def Var(name, type, is_indexed_id=False):
    return Term(name=name, type=type, is_var=True, is_indexed_id=is_indexed_id)


def Const(name, is_indexed_id=False, type="Unknown"):
    return Term(
        name=name, type=type, is_const=True, is_indexed_id=is_indexed_id
    )


def Expr(op, subterms, is_indexed_id=False):
    return Term(op=op, subterms=subterms)


def UnknownSymbol(name):
    return Term(name=name)


def Quantifier(quantifier, quantified_vars, subterms):
    return Term(
        quantifier=quantifier,
        quantified_vars=quantified_vars,
        subterms=subterms
    )


def LetBinding(var_binders, let_terms, subterms):
    return Term(
        var_binders=var_binders,
        let_terms=let_terms,
        subterms=subterms
    )


def LabeledTerm(label, subterms):
    return Term(label=label, subterms=subterms)


class Term:
    def __init__(
        self,
        name=None,
        type=None,
        is_const=None,
        is_var=None,
        label=None,
        indices=None,
        quantifier=None,
        quantified_vars={},
        var_binders=None,
        let_terms=None,
        op=None,
        subterms=None,
        is_indexed_id=False,
        parent=None,
    ):

        self._initialize(
            name=name,
            type=type,
            is_const=is_const,
            is_var=is_var,
            indices=indices,
            label=label,
            quantifier=quantifier,
            quantified_vars=quantified_vars,
            var_binders=var_binders,
            let_terms=let_terms,
            op=op,
            subterms=subterms,
            is_indexed_id=is_indexed_id,
            parent=parent,
        )
        self._add_parent_pointer()

    def _initialize(
        self,
        name=None,
        type=None,
        is_const=None,
        is_var=None,
        label=None,
        indices=None,
        quantifier=None,
        quantified_vars={},
        var_binders=None,
        let_terms=None,
        op=None,
        subterms=None,
        is_indexed_id=None,
        parent=None,
    ):
        self.name = name
        self.type = type
        self.is_const = is_const
        self.is_var = is_var
        self.label = label
        self.indices = indices
        self.quantifier = quantifier
        self.quantified_vars = quantified_vars
        self.var_binders = var_binders
        self.let_terms = let_terms
        self.op = op
        self.subterms = subterms
        self.is_indexed_id = is_indexed_id
        self.parent = parent

    def _add_parent_pointer(self):
        """
        Adds pointer from each element in subterm to expr.
        """
        if self.subterms:
            for term in self.subterms:
                if not isinstance(term, str):
                    term.parent = self

    def find_all(self, e, occs):
        """
        Find all expressions e in self and add them to the list occs.
        """
        if self == e:
            return occs.append(e)
        if self.subterms:
            for sub in self.subterms:
                if sub == e:
                    occs.append(sub)
                else:
                    sub.find_all(e, occs)

    def substitute(self, e, repl):
        """
        Substitute all expressions e in self by repl.
        """
        occs = []
        self.find_all(e, occs)
        for occ in occs:
            occ._initialize(
                name=copy.deepcopy(repl.name),
                type=copy.deepcopy(repl.type),
                is_const=copy.deepcopy(repl.is_const),
                is_var=copy.deepcopy(repl.is_var),
                label=copy.deepcopy(repl.label),
                indices=copy.deepcopy(repl.indices),
                quantifier=copy.deepcopy(repl.quantifier),
                quantified_vars=copy.deepcopy(repl.quantified_vars),
                var_binders=copy.deepcopy(repl.var_binders),
                let_terms=copy.deepcopy(repl.let_terms),
                op=copy.deepcopy(repl.op),
                subterms=copy.deepcopy(repl.subterms),
                is_indexed_id=copy.deepcopy(repl.is_indexed_id),
                parent=occ.parent,
            )

    def __eq__(self, other):
        if not isinstance(other, Term):
            return False
        if self.name != other.name:
            return False
        if self.type != other.type:
            return False
        if self.is_const != other.is_const:
            return False
        if self.is_var != other.is_var:
            return False
        if self.label != other.label:
            return False
        if self.indices != other.indices:
            return False
        if self.quantifier != other.quantifier:
            return False
        if self.quantified_vars != other.quantified_vars:
            return False
        if self.type != other.type:
            return False
        if self.is_var != other.is_var:
            return False
        if self.op != other.op:
            return False
        if self.subterms != other.subterms:
            return False
        if self.is_indexed_id != other.is_indexed_id:
            return False
        return True

    def __get_subterm_str__(self):
        subs_str = ""
        length = len(self.subterms)
        for i in range(length):
            sb = self.subterms[i]
            if i == length - 1:
                subs_str += sb.__str__()
            else:
                subs_str += sb.__str__() + " "
        return subs_str

    def __str__(self):
        if self.is_const or self.is_var or self.is_indexed_id:
            return self.name

        if self.quantifier:
            subs_str = self.__get_subterm_str__()
            n_vars = len(self.quantified_vars[0])
            s = (
                "("
                + self.quantified_vars[0][0]
                + " "
                + self.quantified_vars[1][0]
                + ")"
            )
            if len(self.quantified_vars[0]) > 1:
                for i in range(1, n_vars):
                    s += (
                        " ("
                        + self.quantified_vars[0][i]
                        + " "
                        + self.quantified_vars[1][i]
                        + ")"
                    )
            return "(" + self.quantifier + " (" + s + ") " + subs_str + ")"

        elif self.var_binders:
            s = "(let ("
            for i, var in enumerate(self.var_binders):
                s += "(" + var + " " + self.let_terms[i].__str__() + ")"
            s += ")"

            for sub in self.subterms:
                s += " " + sub.__str__()
            return s + ")"

        elif self.label:
            subs_str = self.__get_subterm_str__()
            return "(! "\
                   + subs_str + " "\
                   + self.label[0] + " " + self.label[1] + ")"
        else:
            subs_str = self.__get_subterm_str__()
            return "(" + self.op.__str__() + " " + subs_str + ")"

    def __repr__(self):
        if self.is_const:
            return self.name

        if self.is_var:
            return self.name + ":" + self.type



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
import copy
import time
import shutil
import random
import signal
import logging
import pathlib

from yinyang.src.core.Statistic import Statistic
from yinyang.src.core.Solver import Solver, SolverQueryResult, SolverResult

from yinyang.src.parsing.Parse import parse_file
from yinyang.src.parsing.Typechecker import typecheck

from yinyang.src.mutators.TypeAwareOpMutation import TypeAwareOpMutation
from yinyang.src.mutators.SemanticFusion.SemanticFusion import SemanticFusion
from yinyang.src.mutators.GenTypeAwareMutation.Util import get_unique_subterms
from yinyang.src.mutators.GenTypeAwareMutation.GenTypeAwareMutation import (
    GenTypeAwareMutation
)


from yinyang.src.base.Utils import random_string, plain, escape
from yinyang.src.base.Exitcodes import OK_BUGS, OK_NOBUGS, ERR_EXHAUSTED_DISK

from yinyang.src.core.Logger import (
    init_logging,
    log_strategy_num_seeds,
    log_generation_attempt,
    log_finished_generations,
    log_crash_trigger,
    log_ignore_list_mutant,
    log_duplicate_trigger,
    log_segfault_trigger,
    log_solver_timeout,
    log_soundness_trigger,
    log_invalid_mutant,
)
from yinyang.src.core.FuzzerUtil import (
    get_seeds,
    grep_result,
    admissible_seed_size,
    in_crash_list,
    in_duplicate_list,
    in_ignore_list,
    init_oracle,
)

MAX_TIMEOUTS = 32


class Fuzzer:
    def __init__(self, args, strategy):
        self.args = args
        self.currentseeds = ""
        self.strategy = strategy
        self.statistic = Statistic()
        self.generator = None
        self.old_time = time.time()
        self.start_time = time.time()
        self.first_status_bar_printed = False
        self.name = random_string()
        self.timeout_of_current_seed = 0

        init_logging(strategy, self.args.quiet, self.name, args)

    def process_seed(self, seed):
        if not admissible_seed_size(seed, self.args):
            self.statistic.invalid_seeds += 1
            logging.debug("Skip invalid seed: exceeds max file size")
            return None, None

        self.currentseeds = pathlib.Path(seed).stem
        script, glob = parse_file(seed, silent=True)

        if not script:

            # Parsing was unsuccessful.
            self.statistic.invalid_seeds += 1
            logging.debug("Skipping invalid seed: error in parsing")
            return None, None

        return script, glob

    def get_script(self, seeds):
        seed = seeds.pop(random.randrange(len(seeds)))
        logging.debug("Processing seed " + seed)
        self.statistic.total_seeds += 1
        return self.process_seed(seed)

    def get_script_pair(self, seeds):
        seed = seeds.pop(random.randrange(len(seeds)))
        seed1 = seed[0]
        seed2 = seed[1]
        logging.debug("Processing seeds " + seed1 + " " + seed2)
        self.statistic.total_seeds += 2
        script1, glob1 = self.process_seed(seed1)
        script2, glob2 = self.process_seed(seed2)
        return script1, glob1, script2, glob2

    def max_timeouts_reached(self):
        if self.timeout_of_current_seed >= MAX_TIMEOUTS:
            return True
        return False  # stop testing if timeout limit is exceeded

    def run(self):
        """
        Realizes the main fuzzing loop. The procedure fetches seeds at random
        from the seed corpus (or a pair of seeds for yinyang), instantiates a
        mutator and then generates `self.args.iterations` many iterations per
        seed.
        """
        seeds = get_seeds(self.args, self.strategy)
        log_strategy_num_seeds(self.strategy, seeds, self.args.SOLVER_CLIS)

        while len(seeds) != 0:
            if self.strategy == "typefuzz":
                script, glob = self.get_script(seeds)
                if not script:
                    continue

                typecheck(script, glob)
                script_cp = copy.deepcopy(script)
                unique_expr = get_unique_subterms(script_cp)
                self.mutator = GenTypeAwareMutation(
                    script, self.args, unique_expr
                )

            elif self.strategy == "opfuzz":
                script, _ = self.get_script(seeds)
                if not script:
                    continue
                self.mutator = TypeAwareOpMutation(script, self.args)

            elif self.strategy == "yinyang":
                script1, _, script2, _ = self.get_script_pair(seeds)
                if not script1 or not script2:
                    continue
                self.mutator = SemanticFusion(script1, script2, self.args)

            else:
                assert False

            log_generation_attempt(self.args)

            unsuccessful_gens = 0
            for i in range(self.args.iterations):
                self.print_stats()
                mutant, success, skip_seed = self.mutator.mutate()

                # Reason for unsuccessful generation: randomness in the
                # mutator to more efficiently generate mutants.
                if not success:
                    self.statistic.unsuccessful_generations += 1
                    unsuccessful_gens += 1
                    continue  # Go to next iteration.

                # Reason for mutator to skip a seed: no random components, i.e.
                # mutant would be the same for all  iterations and hence just
                # waste time.
                if skip_seed:
                    break  # Continue to next seed.

                if not self.test(mutant, i + 1):  # Continue to next seed.
                    break

                self.statistic.mutants += 1

            log_finished_generations(self.args, unsuccessful_gens)
        self.terminate()

    def create_testbook(self, script):
        """
        Generate a "testbook" for script and solver configs.

        script:     parsed SMT-LIB script
        :returns:   list containing with cli and testcases pairs
        """
        testbook = []
        testcase = "%s/%s-%s-%s.smt2" % (
            self.args.scratchfolder,
            escape(self.currentseeds),
            self.name,
            random_string(),
        )
        with open(testcase, "w") as testcase_writer:
            testcase_writer.write(script.__str__())

        for cli in self.args.SOLVER_CLIS:
            testbook.append((cli, testcase))
        return testbook

    def test(self, script, iteration):
        """
        Tests the solvers on the provided script. Checks for crashes, segfaults
        invalid models and soundness issues, ignores duplicates. Stores bug
        triggers in ./bugs along with .output files for bug reproduction.

        script:     parsed SMT-LIB script
        iteration:  number of current iteration (used for logging)
        :returns:   False if the testing on the script should be stopped
                    and True otherwise.
        """

        # For differential testing (opfuzz), the oracle is set to "unknown" and
        # gets overwritten by the result of the first solver call. For
        # metamorphic testing (yinyang) the oracle is pre-set by the cmd line.
        if self.strategy in ["opfuzz", "typefuzz"]:
            oracle = SolverResult(SolverQueryResult.UNKNOWN)
        else:
            oracle = init_oracle(self.args)

        testbook = self.create_testbook(script)
        reference = None

        for testitem in testbook:
            solver_cli, scratchfile = testitem[0], testitem[1]
            solver = Solver(solver_cli)
            self.statistic.solver_calls += 1
            stdout, stderr, exitcode = solver.solve(
                scratchfile, self.args.timeout
            )

            if self.max_timeouts_reached():
                return False

            # Match stdout and stderr against the crash list
            # (see yinyang/config/Config.py:27) which contains various
            # crash messages such as assertion errors, check failure, 
            # invalid models, etc.
            if in_crash_list(stdout, stderr):

                # Match stdout and stderr against the duplicate list
                # (see yinyang/config/Config.py:51) to prevent catching 
                # duplicate bug triggers.
                if not in_duplicate_list(stdout, stderr):
                    self.statistic.effective_calls += 1
                    self.statistic.crashes += 1
                    path = self.report(
                        scratchfile, "crash", solver_cli, stdout, stderr
                    )
                    log_crash_trigger(path)
                else:
                    self.statistic.duplicates += 1
                    log_duplicate_trigger()
                return False  # Stop testing.
            else:

                # Check whether the solver call produced errors, e.g, related
                # to its parser, options, type-checker etc., by matching stdout
                # and stderr against the ignore list 
                # (see yinyang/config/Config.py:54).
                if in_ignore_list(stdout, stderr):
                    log_ignore_list_mutant(solver_cli)
                    self.statistic.invalid_mutants += 1
                    continue  # Continue to the next solver.

                if exitcode != 0:

                    # Check whether the solver crashed with a segfault.
                    if exitcode == -signal.SIGSEGV or exitcode == 245:
                        self.statistic.effective_calls += 1
                        self.statistic.crashes += 1
                        path = self.report(
                            scratchfile, "segfault", solver_cli, stdout, stderr
                        )
                        log_segfault_trigger(self.args, path, iteration)
                        return False  # Stop testing.

                    # Check whether the solver timed out.
                    elif exitcode == 137:
                        self.statistic.timeout += 1
                        self.timeout_of_current_seed += 1
                        log_solver_timeout(self.args, solver_cli, iteration)
                        continue  # Continue to the next solver.

                    # Check whether a "command not found" error occurred.
                    elif exitcode == 127:
                        continue  # Continue to the next solver.

                # Check if the stdout contains a valid solver query result,
                # i.e., contains lines with 'sat', 'unsat' or 'unknown'.
                elif (
                    not re.search("^unsat$", stdout, flags=re.MULTILINE)
                    and not re.search("^sat$", stdout, flags=re.MULTILINE)
                    and not re.search("^unknown$", stdout, flags=re.MULTILINE)
                ):
                    self.statistic.invalid_mutants += 1
                    log_invalid_mutant(self.args, iteration)
                    continue  # Continue to the next solver.

                else:

                    # Grep for '^sat$', '^unsat$', and '^unknown$' to produce
                    # the output (including '^unknown$' to also deal with
                    # incremental benchmarks) for comparing with the oracle
                    # (yinyang) or with other non-erroneous solver runs
                    # (opfuzz) for soundness bugs.
                    self.statistic.effective_calls += 1
                    result = grep_result(stdout)
                    if oracle.equals(SolverQueryResult.UNKNOWN):

                        # For differential testing (opfuzz), the first solver
                        # is set as the reference, its result to be the oracle.
                        oracle = result
                        reference = (solver_cli, scratchfile, stdout, stderr)

                    # Comparing with the oracle (yinyang) or with other
                    # non-erroneous solver runs (opfuzz) for soundness bugs.
                    if not oracle.equals(result):
                        self.statistic.soundness += 1

                        if reference:

                            # Produce a bug report for soundness bugs
                            # containing a diff with the reference solver
                            # (opfuzz).
                            ref_cli = reference[0]
                            ref_stdout = reference[1]
                            ref_stderr = reference[2]
                            path = self.report_diff(
                                scratchfile,
                                "incorrect",
                                ref_cli,
                                ref_stdout,
                                ref_stderr,
                                solver_cli,
                                stdout,
                                stderr,
                            )
                        else:

                            # Produce a bug report if the query result differs
                            # from the pre-set oracle (yinyang).
                            path = self.report(
                                scratchfile, "incorrect", solver_cli,
                                stdout, stderr
                            )

                        log_soundness_trigger(self.args, iteration, path)
                        return False  # Stop testing.
        return True  # Continue to next seed.

    def report(self, scratchfile, bugtype, cli, stdout, stderr):
        plain_cli = plain(cli)
        # format: <solver><{crash,wrong,invalid_model}><seed>.<random-str>.smt2
        report = "%s/%s-%s-%s-%s.smt2" % (
            self.args.bugsfolder,
            bugtype,
            plain_cli,
            escape(self.currentseeds),
            random_string(),
        )
        try:
            shutil.copy(scratchfile, report)
        except Exception:
            logging.error("error: couldn't copy scratchfile to bugfolder.")
            exit(ERR_EXHAUSTED_DISK)
        logpath = "%s/%s-%s-%s-%s.output" % (
            self.args.bugsfolder,
            bugtype,
            plain_cli,
            escape(self.currentseeds),
            random_string(),
        )
        with open(logpath, "w") as log:
            log.write("command: " + cli + "\n")
            log.write("stderr:\n")
            log.write(stderr)
            log.write("stdout:\n")
            log.write(stdout)
        return report

    def report_diff(
        self,
        scratchfile,
        bugtype,
        ref_cli,
        ref_stdout,
        ref_stderr,
        sol_cli,
        sol_stdout,
        sol_stderr,
    ):
        plain_cli = plain(sol_cli)
        # format: <solver><{crash,wrong,invalid_model}><seed>.<random-str>.smt2
        report = "%s/%s-%s-%s-%s.smt2" % (
            self.args.bugsfolder,
            bugtype,
            plain_cli,
            escape(self.currentseeds),
            random_string(),
        )
        try:
            shutil.copy(scratchfile, report)
        except Exception:
            logging.error("error: couldn't copy scratchfile to bugfolder.")
            exit(ERR_EXHAUSTED_DISK)

        logpath = "%s/%s-%s-%s-%s.output" % (
            self.args.bugsfolder,
            bugtype,
            plain_cli,
            escape(self.currentseeds),
            random_string(),
        )
        with open(logpath, "w") as log:
            log.write("*** REFERENCE \n")
            log.write("command: " + ref_cli + "\n")
            log.write("stderr:\n")
            log.write(ref_stderr)
            log.write("stdout:\n")
            log.write(ref_stdout)
            log.write("\n\n*** INCORRECT \n")
            log.write("command: " + sol_cli + "\n")
            log.write("stderr:\n")
            log.write(sol_stderr)
            log.write("stdout:\n")
            log.write(sol_stdout)
        return report

    def print_stats(self):
        if not self.first_status_bar_printed\
           and time.time() - self.old_time >= 1:
            self.statistic.printbar(self.start_time)
            self.old_time = time.time()
            self.first_status_bar_printed = True

        if time.time() - self.old_time >= 2.0:
            self.statistic.printbar(self.start_time)
            self.old_time = time.time()

    def terminate(self):
        print("All seeds processed", flush=True)
        if not self.args.quiet:
            self.statistic.printsum()
        if self.statistic.crashes + self.statistic.soundness == 0:
            exit(OK_NOBUGS)
        exit(OK_BUGS)

    def __del__(self):
        for fn in os.listdir(self.args.scratchfolder):
            if self.name in fn:
                os.remove(os.path.join(self.args.scratchfolder, fn))



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yinyang.src.base.Version import VERSION, COMMIT


def raise_runtime_error(trace, argv, e):
    fn = trace[-1].filename
    lineno = trace[-1].lineno
    print("Runtime error at %s:%s" % (fn, lineno), flush=True)
    print("msg: " + str(e), flush=True)
    print(
        "cmd: "
        + " ".join(argv[:-2])
        + " "
        + '"' + argv[-2]
        + '"' + " " + argv[-1],
        flush=True,
    )
    print("version: yinyang " + VERSION + " " + COMMIT, flush=True)
    print("Please file an issue: https://github.com/testsmt/yinyang/issues",
          flush=True)



# MIT License
#
# Copyright (c) [2020 - 2020] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import logging


class Statistic:
    def __init__(self):
        self.starttime = time.time()
        self.total_seeds = 0
        self.invalid_seeds = 0
        self.total_generations = 0
        self.unsuccessful_generations = 0
        self.mutants = 0
        self.invalid_mutants = 0
        self.crashes = 0
        self.soundness = 0
        self.duplicates = 0
        self.timeout = 0
        self.solver_calls = 0
        self.effective_calls = 0

    def printbar(self, start_time):
        total_time = time.time() - start_time
        if self.solver_calls != 0:
            eff = round(
                (float(self.effective_calls) / float(self.solver_calls)) * 100,
                1
            )
            eff_str = str(eff) + "%"
        else:
            eff_str = "NaN"

        solver_calls_per_sec = round(
            float(self.solver_calls) / float(total_time), 1)

        mutants_per_sec = round(float(self.mutants) / float(total_time), 1)
        mutants_per_sec_str = str(mutants_per_sec)
        bar = "Performed %d solver calls \
(%s calls/s, eff: %s, %s mutants/s)" % (
            self.solver_calls,
            solver_calls_per_sec,
            eff_str,
            mutants_per_sec_str,
        )
        logging.info(bar)

    def printsum(self):
        valid_seeds = self.total_seeds - self.invalid_seeds
        num_bugs = self.crashes + self.soundness
        summary = (
            "\b\b%d seeds processed, \
%d valid, %d invalid \n%d bug triggers found"
            % (self.total_seeds, valid_seeds, self.invalid_seeds, num_bugs)
        )
        print(summary)



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import traceback
import logging

from antlr4.error.ErrorListener import ErrorListener

from yinyang.src.parsing.SMTLIBv2Lexer import SMTLIBv2Lexer
from yinyang.src.parsing.SMTLIBv2Parser import SMTLIBv2Parser
from yinyang.src.parsing.TimeoutDecorator import exit_after
from yinyang.src.parsing.AstVisitor import AstVisitor

from antlr4.CommonTokenStream import CommonTokenStream
from antlr4.FileStream import FileStream
from antlr4.InputStream import InputStream

sys.setrecursionlimit(100000)


class ErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        logging.debug("Parser error on line %d column %d."
                      % (line, column), flush=True)


def prepare_seed(formula):
    """
    Prepare seed script for fuzzing. Remove set-logic, set-info and other
    output-producing commands. Set-logic and set-info may raise warning
    messages from SMT solvers. Output-producing commands may cause errors,
    e.g., get-model, get-proof etc. Errors such as

                (error "line X column Y: msg")

    are ignored to avoid false positives in the bug detection logic. As the bug
    detection logic is based on string matching such errors may lead to
    soundness issues being ignored, e.g. if the error occurred after a faulty
    check-sat result. Hence, we remove all output-producing SMT-LIB commands
    from the script.
    """
    new_cmds = []
    for cmd in formula.commands:
        if "set-info" in cmd.__str__():
            continue
        if "set-logic" in cmd.__str__():
            continue

        # Ignore output-producing commands to make sure the detection logic
        # won't be mislead
        #
        if "get-model" in cmd.__str__():
            continue
        if "get-assertions" in cmd.__str__():
            continue
        if "get-proof" in cmd.__str__():
            continue
        if "get-unsat-assumptions" in cmd.__str__():
            continue
        if "get-unsat-core" in cmd.__str__():
            continue
        if "get-value" in cmd.__str__():
            continue
        if "echo" in cmd.__str__():
            continue
        if "simplify" in cmd.__str__():
            continue
        new_cmds.append(cmd)

    formula.commands = new_cmds
    return formula


def generate_ast(stream, prep_seed=True):
    error_listener = ErrorListener()
    lexer = SMTLIBv2Lexer(stream)
    lexer.removeErrorListeners()
    lexer.addErrorListener(error_listener)
    stream = CommonTokenStream(lexer)
    parser = SMTLIBv2Parser(stream)
    parser.removeErrorListeners()
    tree = parser.start()
    vis = AstVisitor()
    formula = vis.visitStart(tree)

    # empty file or parser preceding parser error
    if len(formula.commands) == 0:
        return None

    return prepare_seed(formula) if prep_seed else formula, vis.global_vars


def parse_filestream(fn, timeout_limit):
    @exit_after(timeout_limit)
    def _parse_filestream(fn):
        fstream = FileStream(fn, encoding="utf8")
        ast, globs = generate_ast(fstream)
        return ast, globs

    return _parse_filestream(fn)


def parse_inputstream(s, timeout_limit):
    @exit_after(timeout_limit)
    def _parse_inputstream(s):
        istream = InputStream(s)
        ast, globs = generate_ast(istream)
        return ast, globs

    return _parse_inputstream(s)


def parse(parse_fct, arg, timeout_limit, silent=True):
    """
    Parser helper function.

    :parse_fct: function to parse stream.
    :arg: first argument to parse_fct.
    :returns: Script object representing AST of SMT-LIB file. None if timeout
              or crash occurred.
    """
    script = None
    globs = None

    try:
        script, globs = parse_fct(arg, timeout_limit)
    except KeyboardInterrupt:
        print("Parser timed out or was interrupted.")
    except Exception as e:
        if not silent:
            print("Error generating the AST.")
            print(e)
            traceback.print_exc(file=sys.stdout)
    return script, globs


def parse_file(fn, timeout_limit=30, silent=True):
    """
    Parse SMT-LIB file.

    :fn: path to SMT-LIB file.
    :silent: if silent=True the parser will withhold stacktrace from user
             on crash.
    :returns: Script object representing AST of SMT-LIB file. None if timeout
              or crash occurred.
    """
    return parse(parse_filestream, fn, timeout_limit, silent)


def parse_str(s, timeout_limit=30, silent=True):
    """
    Parse SMT-LIB from string.

    :fn: path to SMT-LIB file.
    :silent: if silent=True the parser will withhold stacktrace from user
             on crash.
    :returns: Script object representing AST of SMT-LIB file. None if timeout
              or crash occurred.
    """
    return parse(parse_inputstream, s, timeout_limit, silent)



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import itertools

from yinyang.src.parsing.Ast import (
    Term,
    Script,
    Assert,
    DeclareConst,
    DeclareFun,
    SMTLIBCommand,
)


def cvars(occs):
    """
    Return a single representative occurrence for each variable.
    """
    names = []
    canonicals = []
    for occ in occs:
        if occ.name in names:
            continue
        canonicals.append(occ)
    return canonicals


def debug_formula(formula, name="formula"):
    print("#" * 10, name, "#" * 10)
    print(formula.__str__())
    print("#" * (10 + len(name) + 10))
    print()


def is_constant(cmd):
    if isinstance(cmd, DeclareConst):
        return True
    if isinstance(cmd, DeclareFun) and cmd.input_sort == "":
        return True
    return False


def is_sort(cmd):
    if isinstance(cmd, SMTLIBCommand) and "-sort" in cmd.cmd_str:
        return True
    return False


def concat(op, script1, script2):
    script1.merge_asserts()
    script2.merge_asserts()
    sorts = []
    sorts = [
        cmd for cmd in script1.commands + script2.commands if is_sort(cmd)
    ]
    sorts = list(set(sorts))
    declares1 = [cmd for cmd in script1.commands if is_constant(cmd)]
    assert1 = [cmd for cmd in script1.commands if isinstance(cmd, Assert)][0]
    assert2 = [cmd for cmd in script2.commands if isinstance(cmd, Assert)][0]
    conjunction = Assert(Term(op=op, subterms=[assert1.term, assert2.term]))
    new_cmds = declares1

    for cmd in script2.commands:
        if is_sort(cmd):
            continue
        if isinstance(cmd, Assert):
            new_cmds.append(conjunction)
        else:
            new_cmds.append(cmd)
    new_cmds = sorts + new_cmds
    return Script(new_cmds, {**script1.global_vars, **script2.global_vars})


def conjunction(script1, script2):
    return concat("and", script1, script2)


def disjunction(script1, script2):
    return concat("or", script1, script2)


def type_var_map(global_vars):
    mapping = {}
    for var in global_vars:
        if global_vars[var] not in mapping:
            mapping[global_vars[var]] = [var]
        else:
            if var not in mapping[global_vars[var]]:
                mapping[global_vars[var]].append(var)
    return mapping


def random_tuple_list(lst1, lst2, lb=1):
    """
    Generate a random list of tuples (x,y) where x is in lst1 and y is in lst2;
    """
    product = list(itertools.product(lst1, lst2))

    if len(product) == 0:
        k = 0
    else:
        k = random.randint(lb, len(product))
    tups = random.sample(product, k)
    random.shuffle(tups)

    new_tups = []
    lhs, rhs = [], []
    for tup in tups:
        if tup[0] in lhs:
            continue
        if tup[1] in rhs:
            continue
        lhs.append(tup[0])
        rhs.append(tup[1])
        new_tups.append(tup)
    return new_tups


def random_var_triplets(global_vars1, global_vars2, templates):
    """
    Create a random variable mapping of variables with same type
    """
    m1, m2 = type_var_map(global_vars1), type_var_map(global_vars2)
    mapping = []
    for t in templates:
        if t not in m1:
            continue
        if t not in m2:
            continue
        random_tuples = random_tuple_list(m1[t], m2[t])
        for tup in random_tuples:
            mapping.append((tup[0], tup[1], random.choice(templates[t]), t))
    return mapping



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yinyang.src.base.Version import VERSION


header = (
    "opfuzz -- a differential fuzzer for SMT solvers [version: "
    + VERSION
    + "]"
)

usage = """ opfuzz [options] solver_clis seed_file   [optionally, more seed files]
       opfuzz [options] solver_clis seed_folder [optionally, more seed folders]
       solver_clis := "solver_cli1;solver_cli2;...;solver_clik"
"""

short_description = """
opfuzz is a tool for finding bugs in SMT solver by generating mutant SMT-LIB
scripts from a set of seed SMT-LIB scripts (.smt2 files) and then cross-check
the results of the solvers. opfuzz realizes type-aware operator mutations.

Example (basic usage):
$ opfuzz "z3 model_validate=true;cvc5 --check-models -m -i -q" benchmarks
[2021/06/10 12:12:40 PM] Strategy: opfuzz, 2 testing targets, 299476 seeds
[2021/06/10 12:12:41 PM] Performed 36 solver calls (35.1 calls/s, eff: 88.9%, 17.6 mutants/s)
[2021/06/10 12:12:46 PM] Performed 254 solver calls (41.9 calls/s, eff: 95.3%, 21.0 mutants/s)
User interrupt
1 seeds processed, 1 valid, 0 invalid
0 bug triggers found

This will by default randomly select smt2 files from the folder `./benchmarks`
and generate 300 mutants per file. If a bug has been found, opfuzz will store
it in `./bugs`. You can use the shortcut CTRL+C to terminate opfuzz manually.
For a listing of options and further information, use opfuzz --help."""   # noqa: E501

long_description = """
opfuzz is a mutation-based fuzzer, i.e. it mutates a set of seed formulas using
type-aware operator mutations and then uses the mutated formulas as the test
seeds for SMT solvers. It can so detect soundness bugs, invalid model bugs,
crashes, segfaults, etc. opfuzz's mutation strategy generates mutants by
interchanging operators, e.g, =, distinct, +, -,  *, / by one another. opfuzz
is based on differential testing, i.e. it needs at least two SMT solvers to
compare their results.


*** Example 1 (basic usage) ***

The following command runs opfuzz on the two SMT solver z3 and cvc5 with seed
formulas from the folder "benchmarks" and shows its expected behavior.

    $ opfuzz "z3 model_validate=true;cvc5 --check-models -m -i -q" benchmarks
    [2021/06/10 11:07:39 AM] Strategy: opfuzz, 2 testing targets, 299476 seeds
    [2021/06/10 11:07:40 AM] Performed 108 solver calls (107.5 calls/s, eff: 100.0%, 53.8 mutants/s)
    [2021/06/10 11:07:45 AM] Performed 630 solver calls (104.5 calls/s, eff: 98.1%, 52.2 mutants/s)
    [2021/06/10 11:07:50 AM] Performed 782 solver calls (70.6 calls/s, eff: 89.0%, 35.3 mutants/s)
    [2021/06/10 11:07:55 AM] Performed 936 solver calls (58.2 calls/s, eff: 82.6%, 29.1 mutants/s)
    [2021/06/10 11:08:00 AM] Performed 1090 solver calls (51.6 calls/s, eff: 78.3%, 25.8 mutants/s)
    [2021/06/10 11:08:20 AM] Performed 1202 solver calls (29.5 calls/s, eff: 76.3%, 14.8 mutants/s)
    [2021/06/10 11:08:30 AM] Performed 1204 solver calls (23.8 calls/s, eff: 76.2%, 11.9 mutants/s)
    User interrupt
    3 seeds processed, 3 valid, 0 invalid
    0 bug triggers found

opfuzz randomly selects the seed formulas from the folder and, by default,
ignores files larger than 100k. opfuzz will generate 300 mutants per seed
formula, with which it tests the SMT solvers. You can terminate opfuzz by CTLR + C.
opfuzz stores bug trigger in ./bugs and creates rolling logs up to 5MB per run
in ./logs. Note, if you run opfuzz many times in a row, it might be worthwhile
to disable logging to avoid an overflow of the logs directory.

Every two seconds, opfuzz will generate statistics on (1) solver calls,
(2) efficiency and (3) mutants per second. Efficiency corresponds to the percentage
of solver calls that did not timeout, cause parser or unsupported option errors.
It is expected that efficiency, solver calls and mutants per second can vary
quite a bit during a regular run of opfuzz. However, if the efficiency is
very low (e.g., < 50%), it may be a good idea to inspect the logs, see what
is going on and potentially adjust seed set and solver commandlines. Note, some
SMT solvers need options to support a particular set of seeds. E.g, cvc5 needs
the option "--strings-exp" to support string logic. Otherwise it will often
throw an error message.

*** Example 2 (finding a bug with opfuzz) ***
The following command sequence will let you find a soundness bug in cvc4-1.7.

** 1. Get cvc4-1.7 **
For linux, execute:

    $ wget https://github.com/cvc5/cvc5/releases/download/1.7/cvc4-1.7-x86_64-linux-opt
    $ chmod + x cvc4-1.7-x86_64-linux-opt

For OSX and windows, you can build cvc4 from tarball available on GitHub:
https://github.com/cvc5/cvc5/releases/tag/1.7.

** 2. Seed **
Create a file seed.smt2 with the following content

    $ cat seed.smt2
    (assert (and (< 60.3 (exp 4.1) 60.4) (< (exp 5.1) 164.1)))
    (check-sat)

** 3. Run opfuzz **
Now, run opfuzz with the following configurations on the fabricated seed.

    $ opfuzz "./cvc4-1.7-x86_64-linux-opt -q --sygus-inference;./cvc4-1.7-x86_64-linux-opt -q" seed.smt2
    [2021/06/10 02:47:49 PM] Strategy: opfuzz, 2 testing targets, 1 seeds
    [2021/06/10 02:47:49 PM] Detected soundness bug! bugs/incorrect-cvc4-17-x86_64-linux-opt-q-seed-5yCdH.smt2
    All seeds processed
    1 seeds processed, 1 valid, 0 invalid
    1 bug triggers found

You can now observe the bug by executing

    $ ./cvc4-1.7-x86_64-linux-opt -q --sygus-inference bugs/incorrect-cvc4-17-x86_64-linux-opt-q-seed-5yCdH.smt2
    unsat
    $ ./cvc4-1.7-x86_64-linux-opt -q bugs/incorrect-cvc4-17-x86_64-linux-opt-q-seed-5yCdH.smt2
    sat
"""  # noqa: E501


options = """
options:
    -i <N>,  --iterations <N>
            iterations per seed (default: 300)
    -m <N>,  --modulo <N>
            the number of times the seed will be forwarded to the solvers
            For example, with 300 iterations and 2 as a modulo, 150 mutants
            per seed file will be passed to the SMT solvers. (default: 2)
    -l <path>, --logfolder <path>
            log folder (default: ./logs)
    -t <secs>, --timeout <secs>
            timeout per SMT solver call in seconds (default: 8)
    -b <path>, --bugsfolder <path>
            set bug folder (default: ./bugs)
    -s <path>, --scratchfolder <path>
            temp folder to dump mutants. (default: ./scratch)
    -c <file>, --config <file>
            set custom operator mutation config file
    -L <bytes>, --limit <bytes>
            file size limit on seed formula in bytes (default: 100000)
    -n, --no-log    disable logging
    -q, --quiet     do not print statistics and other output
    -v, --version   show version number and exit
    -h, --help      show this help message and exit
"""
